# Generated by Django 4.0.3 on 2022-03-24 01:47

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('reviewpanel', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='input',
            options={'ordering': ['form', '_rank']},
        ),
        migrations.AddField(
            model_name='cohort',
            name='allow_skip',
            field=models.BooleanField(default=True),
        ),
    ]



from django.db import models
from django.db.models import UniqueConstraint, Subquery, OuterRef, Count, \
    Avg, StdDev, Max, Min, Func, BooleanField
from django.db.models.functions import Coalesce, Cast
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import GenericForeignKey, \
    GenericRelation
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from formative.models import Program, Form, RankedModel
from formative.utils import MarkdownFormatter, remove_p


markdown = MarkdownFormatter()


class Template(models.Model):
    class Meta:
        constraints = [
            UniqueConstraint(fields=['program', 'name'],
                             name='unique_program_slug')
        ]
    
    program = models.ForeignKey(Program, models.CASCADE,
                                related_name='templates',
                                related_query_name='template')
    name = models.SlugField(max_length=32, allow_unicode=True,
                            verbose_name='identifier')
    min_width = models.PositiveIntegerField(default=0)
    max_width = models.PositiveIntegerField(default=0)
    
    def __str__(self):
        return self.name


class TemplateSection(models.Model):
    class Meta:
        constraints = [
            UniqueConstraint(fields=['template', 'name'],
                             name='unique_template_name')
        ]
    
    template = models.ForeignKey(Template, models.CASCADE,
                                 related_name='sections',
                                 related_query_name='section')
    name = models.SlugField(max_length=32, allow_unicode=32,
                            verbose_name='identifier')
    x = models.DecimalField(max_digits=7, decimal_places=3, default='0.000',
                            verbose_name='left %')
    y = models.DecimalField(max_digits=7, decimal_places=3, default='0.000',
                            verbose_name='top %')
    w = models.DecimalField(max_digits=7, decimal_places=3, default='100.000',
                            verbose_name='width %')
    h = models.DecimalField(max_digits=7, decimal_places=3,
                            null=True, blank=True, verbose_name='height')
    wrap = models.BooleanField(default=True)
    scroll = models.BooleanField(default=False) # vert, or horizontal if no wrap
    font = models.CharField(max_length=64, blank=True,
                            default='100% sans-serif')
    
    def __str__(self):
        return self.name


class Presentation(models.Model):
    class Meta:
        constraints = [
            UniqueConstraint(fields=['form', 'name'], name='unique_form_slug')
        ]
    
    form = models.ForeignKey(Form, models.CASCADE,
                             related_name='presentations',
                             related_query_name='presentation')
    template = models.ForeignKey(Template, models.CASCADE,
                                 null=True, blank=True,
                                 related_name='presentations',
                                 related_query_name='presentation')
    name = models.SlugField(max_length=32, allow_unicode=True,
                            verbose_name='identifier')
    no_input = models.BooleanField(default=False)
    min_seconds = models.PositiveIntegerField(default=0)
    options = models.JSONField(default=dict, blank=True)
    created = models.DateTimeField(auto_now_add=True)
    
    def __str__(self):
        return self.name
    
    def get_absolute_url(self):
        form = self.form
        args = {'program_slug': form.program.slug, 'form_slug': form.slug,
                'presentation': self.id}
        return reverse('plugins:reviewpanel:presentation_admin', kwargs=args)
    
    def inputs_section(self):
        if self.no_input: return None
        return 'inputs'
    
    def metrics_section(self):
        return 'metrics'
    
    def custom_css(self):
        if 'custom_css' in self.options: return self.options['custom_css']
        return ''
    
    def show_stats(self):
        if 'hide_stats' in self.options: return False
        return True


class Reference(RankedModel):
    class Meta:
        constraints = [
            UniqueConstraint(fields=['presentation', '_rank'],
                             name='unique_presentation_rank')
        ]
        ordering = ['presentation', '_rank']
    
    presentation = models.ForeignKey(Presentation, models.CASCADE,
                                     related_name='references',
                                     related_query_name='reference')
    # if no collection, name identifies the (stock or custom) block
    collection = models.SlugField(max_length=32, allow_unicode=True, blank=True)
    # if no name, reference is to the item file for the specified collection
    name = models.SlugField(max_length=32, allow_unicode=True, blank=True)
    # if a composite block, can also give a field name (e.g., state for address)
    field = models.CharField(max_length=32, blank=True)
    
    section = models.ForeignKey(TemplateSection, models.SET_NULL,
                                null=True, blank=True, related_name='+')
    block_label = models.CharField(max_length=256, blank=True) # in markdown
    inline_label = models.CharField(max_length=128, blank=True)
    options = models.JSONField(default=dict, blank=True)
    
    # fields used when a collection is specified:
    block_combine = models.BooleanField(default=False)
    inline_combine = models.CharField(max_length=16, blank=True) # separator
    # if not combined, and no select section specified, it's added to the bottom
    select_section = models.ForeignKey(TemplateSection, models.SET_NULL,
                                       null=True, blank=True, related_name='+')
    
    def __str__(self):
        if self.collection:
            if self.name: return ' '.join((self.collection, self.name))
            return self.collection + ' file'
        if self.field: return ' '.join((self.name, self.field))
        return self.name
    
    def rank_group(self):
        return Reference.objects.filter(presentation=self.presentation)
    
    @property
    def is_file(self):
        return self.collection and not self.name
    
    @property
    def combined(self):
        if self.block_combine or self.inline_combine: return True
        return False
    
    def id_string(self):
        str = ''
        if self.collection: str += self.collection + '_'
        if self.name: str += self.name
        else: str += 'file'
        
        return str
    
    def block_label_html(self):
        return mark_safe(markdown.convert(self.block_label))
    
    def inline_label_html(self):
        return mark_safe(remove_p(markdown.convert(self.inline_label)))


class Input(RankedModel):
    class Meta:
        constraints = [
            UniqueConstraint(fields=['form', '_rank'], name='unique_form_rank'),
            UniqueConstraint(fields=['form', 'name'], name='unique_input_name')
        ]
        ordering = ['form', '_rank']
    
    MAX_TEXT_MAXLEN = 1000
    
    class InputType(models.TextChoices):
        NUMERIC = 'num', _('numeric')
        BOOLEAN = 'bool', _('true/false')
        TEXT = 'text', _('text')
    
    form = models.ForeignKey(Form, models.CASCADE,
                             related_name='inputs', related_query_name='input')
    name = models.SlugField(max_length=32, allow_unicode=True)
    label = models.CharField(max_length=32, blank=True)
    type = models.CharField(max_length=16, choices=InputType.choices,
                            default=InputType.NUMERIC)
    min_num = models.PositiveIntegerField(null=True, blank=True)
    max_num = models.PositiveIntegerField(null=True, blank=True)
    max_chars = models.PositiveIntegerField(null=True, blank=True)
    
    def __str__(self):
        return self.name
    
    def rank_group(self):
        return Input.objects.filter(form=self.form)


class Metric(models.Model):
    class Meta:
        constraints = [
            UniqueConstraint(fields=['input', 'name'],
                             name='unique_metric_name')
        ]
        ordering = ['input', 'position', 'pk']
    
    class MetricType(models.TextChoices):
        COUNT = 'count', _('count')
        AVG = 'avg', _('average')
        STDDEV = 'stddev', _('standard deviation')
        MAX = 'max', _('maximum')
        MIN = 'min', _('minimum')
    
    input = models.ForeignKey(Input, models.CASCADE,
                              related_name='inputs', related_query_name='input')
    name = models.SlugField(max_length=20, allow_unicode=True)
    type = models.CharField(max_length=16, choices=MetricType.choices,
                            default=MetricType.COUNT)
    cohort = models.ForeignKey('Cohort', models.CASCADE, null=True, blank=True,
                               related_name='+')
    count_value = models.PositiveIntegerField(null=True, blank=True)
    count_is_divisor = models.BooleanField(default=False)
    boolean_invert = models.BooleanField(default=False)
    position = models.PositiveIntegerField(default=1)
    admin_enabled = models.BooleanField(default=True)
    panelist_enabled = models.BooleanField(default=False)
    display_values = models.BooleanField(default=False)
    
    def __str__(self):
        return self.name
    
    def aggregate(self):
        if self.type == self.MetricType.COUNT: return Count
        elif self.type == self.MetricType.AVG: return Avg
        elif self.type == self.MetricType.STDDEV: return StdDev
        elif self.type == self.MetricType.MAX: return Max
        elif self.type == self.MetricType.MIN: return Min
    
    def annotation(self, queryset, **kwargs):
        qs = queryset.filter(input=self.input,
                             **{ k: OuterRef(kwargs[k]) for k in kwargs })
        if self.cohort: qs = qs.filter(cohort=self.cohort)
        
        if self.type == self.MetricType.COUNT and self.count_value is not None:
            qs = qs.filter(value=self.count_value)
        elif self.input.type != Input.InputType.BOOLEAN:
            qs = qs.filter(value__gt=0)
        
        class Not(Func):
            function = 'NOT'
            template = '%(function)s %(expressions)s'
        
        qs = qs.values(*kwargs.keys())
        val = self.aggregate()('value')
        if self.input.type == Input.InputType.BOOLEAN:
            if self.type in (self.MetricType.MAX, self.MetricType.MIN):
                val = Cast(val, output_field=BooleanField())
                if self.boolean_invert: val = Not(val)
        expr = Subquery(qs.annotate(v=val).values('v'))
        
        if self.type == self.MetricType.COUNT: expr = Coalesce(expr, 0)
        return expr


class Panel(models.Model):
    class Meta:
        constraints = [
            UniqueConstraint(fields=['program', 'name'],
                             name='unique_panel_name')
        ]
    
    program = models.ForeignKey(Program, models.CASCADE, related_name='panels',
                                related_query_name='panel')
    name = models.CharField(max_length=50)
    panelists = models.ManyToManyField(User, blank=True, related_name='panels',
                                       related_query_name='panel')
    created = models.DateTimeField(auto_now_add=True)
    
    def __str__(self):
        return self.name


class Cohort(models.Model):
    class Meta:
        constraints = [
            UniqueConstraint(fields=['form', 'name'], name='unique_cohort_name')
        ]
    
    class Status(models.TextChoices):
        INACTIVE = 'inactive', _('inactive')
        ACTIVE = 'active', _('active')
        COMPLETED = 'completed', _('completed')
    
    form = models.ForeignKey(Form, models.CASCADE, # could remove and use
                             related_name='cohorts', # presentation.form?
                             related_query_name='cohort')
    name = models.CharField(max_length=50)
    message = models.TextField(blank=True)
    presentation = models.ForeignKey(Presentation, models.SET_NULL,
                                     null=True, blank=True,
                                     related_name='cohorts',
                                     related_query_name='cohort')
    panel = models.ForeignKey(Panel, models.SET_NULL,
                               null=True, blank=True,
                               related_name='cohorts',
                               related_query_name='cohort')
    status = models.CharField(max_length=16, default=Status.INACTIVE,
                              choices=Status.choices)
    size = models.PositiveIntegerField(default=0, editable=False)
    panel_weight = models.FloatField(default=1.0)
    inputs = models.ManyToManyField(Input, blank=True, related_name='cohorts',
                                    related_query_name='cohort')
    allow_skip = models.BooleanField(default=True)
    created = models.DateTimeField(auto_now_add=True)
    activated = models.DateTimeField(null=True, blank=True, editable=False)
    completed = models.DateTimeField(null=True, blank=True, editable=False)
    
    def __str__(self):
        return self.name
    
    def members(self):
        cohort_members = Subquery(self.cohortmember_set.values('object_id'))
        return self.form.model.objects.filter(pk__in=cohort_members)


class CohortMember(models.Model):
    cohort = models.ForeignKey(Cohort, models.CASCADE)
    content_type = models.ForeignKey(ContentType, models.CASCADE)
    object_id = models.UUIDField(db_index=True)
    member = GenericForeignKey()
    
    def __str__(self):
        if self.cohort.form.validation_type == Form.Validation.EMAIL:
            return self.member._email
        
        return self.object_id


class ScoreManager(models.Manager):
    def create_for_cohort(self, user, cohort, **extra_fields):
        return self.model(panelist=user, form=cohort.form, cohort=cohort,
                          **extra_fields)


class Score(models.Model):
    class Meta:
        constraints = [
            UniqueConstraint(fields=['panelist', 'object_id',
                                     'cohort', 'input'],
                             name='unique_panelist_submission_cohort_input')
        ]
    
    panelist = models.ForeignKey(User, models.SET_NULL, null=True, blank=True,
                                 related_name='scores',
                                 related_query_name='score')
    form = models.ForeignKey(Form, models.CASCADE, related_name='scores',
                             related_query_name='score')
    content_type = models.ForeignKey(ContentType, models.CASCADE)
    object_id = models.UUIDField(db_index=True)
    submission = GenericForeignKey()
    cohort = models.ForeignKey(Cohort, models.CASCADE, related_name='scores',
                               related_query_name='score')
    input = models.ForeignKey(Input, models.CASCADE, related_name='scores',
                              related_query_name='score')
    value = models.PositiveIntegerField(null=True, blank=True)
    text = models.CharField(max_length=Input.MAX_TEXT_MAXLEN, blank=True)
    created = models.DateTimeField(auto_now_add=True)
    
    objects = ScoreManager()
    
    def __str__(self):
        name = f'{self.panelist.username} {self.input.name} '
        if self.value is None: return name + 'placeholder'
        elif not self.value: return name + 'skip'
        elif self.input.type == Input.InputType.TEXT: return name + 'score'
        return name + f'score={self.value}'



# Generated by Django 4.0.3 on 2022-03-29 02:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('reviewpanel', '0004_cohort_size'),
    ]

    operations = [
        migrations.AddField(
            model_name='presentation',
            name='options',
            field=models.JSONField(blank=True, default=dict),
        ),
    ]



# Generated by Django 4.0.3 on 2022-03-23 18:04

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('formative', '0006_alter_formblock__rank'),
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Cohort',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50)),
                ('message', models.TextField(blank=True)),
                ('status', models.CharField(choices=[('inactive', 'inactive'), ('active', 'active'), ('completed', 'completed')], default='inactive', max_length=16)),
                ('panel_weight', models.FloatField(default=1.0)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('activated', models.DateTimeField(blank=True, editable=False, null=True)),
                ('completed', models.DateTimeField(blank=True, editable=False, null=True)),
                ('form', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='cohorts', related_query_name='cohort', to='formative.form')),
            ],
        ),
        migrations.CreateModel(
            name='Input',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('_rank', models.IntegerField(verbose_name='')),
                ('name', models.SlugField(allow_unicode=True, max_length=32)),
                ('label', models.CharField(blank=True, max_length=32)),
                ('type', models.CharField(choices=[('num', 'numeric'), ('bool', 'true/false'), ('text', 'text')], default='num', max_length=16)),
                ('min_num', models.PositiveIntegerField(blank=True, null=True)),
                ('max_num', models.PositiveIntegerField(blank=True, null=True)),
                ('max_chars', models.PositiveIntegerField(blank=True, null=True)),
                ('form', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='inputs', related_query_name='input', to='formative.form')),
            ],
        ),
        migrations.CreateModel(
            name='Presentation',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.SlugField(allow_unicode=True, max_length=32, verbose_name='identifier')),
                ('no_input', models.BooleanField(default=False)),
                ('min_seconds', models.PositiveIntegerField(default=0)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('form', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='presentations', related_query_name='presentation', to='formative.form')),
            ],
        ),
        migrations.CreateModel(
            name='Template',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.SlugField(allow_unicode=True, max_length=32, verbose_name='identifier')),
                ('min_width', models.PositiveIntegerField(default=0)),
                ('max_width', models.PositiveIntegerField(default=0)),
                ('program', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='templates', related_query_name='template', to='formative.program')),
            ],
        ),
        migrations.CreateModel(
            name='TemplateSection',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.SlugField(allow_unicode=32, max_length=32, verbose_name='identifier')),
                ('x', models.DecimalField(decimal_places=3, default='0.000', max_digits=7, verbose_name='left %')),
                ('y', models.DecimalField(decimal_places=3, default='0.000', max_digits=7, verbose_name='top %')),
                ('w', models.DecimalField(decimal_places=3, default='100.000', max_digits=7, verbose_name='width %')),
                ('h', models.DecimalField(blank=True, decimal_places=3, max_digits=7, null=True, verbose_name='height')),
                ('wrap', models.BooleanField(default=True)),
                ('scroll', models.BooleanField(default=True)),
                ('font', models.CharField(blank=True, default='100% sans-serif', max_length=64)),
                ('template', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='sections', related_query_name='section', to='reviewpanel.template')),
            ],
        ),
        migrations.CreateModel(
            name='Score',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.UUIDField()),
                ('value', models.PositiveIntegerField(blank=True, null=True)),
                ('text', models.CharField(blank=True, max_length=1000)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('cohort', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='scores', related_query_name='score', to='reviewpanel.cohort')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
                ('form', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='scores', related_query_name='score', to='formative.form')),
                ('input', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='scores', related_query_name='score', to='reviewpanel.input')),
                ('panelist', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='scores', related_query_name='score', to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='Reference',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('_rank', models.IntegerField(verbose_name='')),
                ('collection', models.SlugField(allow_unicode=True, blank=True, max_length=32)),
                ('name', models.SlugField(allow_unicode=True, blank=True, max_length=32)),
                ('field', models.CharField(blank=True, max_length=32)),
                ('block_label', models.CharField(blank=True, max_length=256)),
                ('inline_label', models.CharField(blank=True, max_length=128)),
                ('options', models.JSONField(blank=True, default=dict)),
                ('block_combine', models.BooleanField(default=False)),
                ('inline_combine', models.CharField(blank=True, max_length=16)),
                ('presentation', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='references', related_query_name='reference', to='reviewpanel.presentation')),
                ('section', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='reviewpanel.templatesection')),
                ('select_section', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='reviewpanel.templatesection')),
            ],
            options={
                'ordering': ['presentation', '_rank'],
            },
        ),
        migrations.AddField(
            model_name='presentation',
            name='template',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='presentations', related_query_name='presentation', to='reviewpanel.template'),
        ),
        migrations.CreateModel(
            name='Panel',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=50)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('panelists', models.ManyToManyField(related_name='panels', related_query_name='panel', to=settings.AUTH_USER_MODEL)),
                ('program', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='panels', related_query_name='panel', to='formative.program')),
            ],
        ),
        migrations.CreateModel(
            name='CohortMember',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.UUIDField()),
                ('cohort', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='reviewpanel.cohort')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
            ],
        ),
        migrations.AddField(
            model_name='cohort',
            name='inputs',
            field=models.ManyToManyField(related_name='cohorts', related_query_name='cohort', to='reviewpanel.input'),
        ),
        migrations.AddField(
            model_name='cohort',
            name='panel',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='cohorts', related_query_name='cohort', to='reviewpanel.panel'),
        ),
        migrations.AddField(
            model_name='cohort',
            name='presentation',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='cohorts', related_query_name='cohort', to='reviewpanel.presentation'),
        ),
        migrations.AddConstraint(
            model_name='templatesection',
            constraint=models.UniqueConstraint(fields=('template', 'name'), name='unique_template_name'),
        ),
        migrations.AddConstraint(
            model_name='template',
            constraint=models.UniqueConstraint(fields=('program', 'name'), name='unique_program_slug'),
        ),
        migrations.AddConstraint(
            model_name='score',
            constraint=models.UniqueConstraint(fields=('panelist', 'object_id', 'cohort', 'input'), name='unique_panelist_submission_cohort_input'),
        ),
        migrations.AddConstraint(
            model_name='reference',
            constraint=models.UniqueConstraint(fields=('presentation', '_rank'), name='unique_presentation_rank'),
        ),
        migrations.AddConstraint(
            model_name='presentation',
            constraint=models.UniqueConstraint(fields=('form', 'name'), name='unique_form_slug'),
        ),
        migrations.AddConstraint(
            model_name='panel',
            constraint=models.UniqueConstraint(fields=('program', 'name'), name='unique_panel_name'),
        ),
        migrations.AddConstraint(
            model_name='input',
            constraint=models.UniqueConstraint(fields=('form', '_rank'), name='unique_form_rank'),
        ),
        migrations.AddConstraint(
            model_name='input',
            constraint=models.UniqueConstraint(fields=('form', 'name'), name='unique_input_name'),
        ),
        migrations.AddConstraint(
            model_name='cohort',
            constraint=models.UniqueConstraint(fields=('form', 'name'), name='unique_cohort_name'),
        ),
    ]



from django import forms
from django.contrib.admin import widgets

from formative.forms import AdminJSONForm, ExportAdminForm, NegatedBooleanField
from formative.models import FormBlock
from .models import TemplateSection, Presentation, Input, Metric, Cohort


class ReferenceForm(AdminJSONForm):
    class Meta:
        exclude = ('_rank',)
        json_fields = {'options': []}
        labels = {
            'section': 'template section',
            'select_section': 'section for selector'
        }
        help_texts = {
            'collection': 'Leave blank to reference a block.',
            'name': "Name of the block, or of the item field. " \
                    "Leave blank to reference the collection's file.",
            'select_section': 'If a separate section is not used, the ' \
                              'selector will be added below the content.'
        }
    
    def __init__(self, presentation=None, extra=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.presentation = presentation
        
        if extra or self.instance and not self.instance.pk:
            # these are set after the instance gets added, when applicable
            for n in ('field', 'block_combine', 'inline_combine',
                      'select_section'):
                self.fields[n].widget = forms.HiddenInput()
        else:
            self.fields['collection'].disabled = True
            
            if self.instance.combined or not self.instance.collection:
                self.fields['select_section'].widget = forms.HiddenInput()
            if not self.instance.collection:
                self.fields['block_combine'].widget = forms.HiddenInput()
                self.fields['inline_combine'].widget = forms.HiddenInput()
                
                form = presentation.form
                self.fields['name'].choices = [ (n, n) for n in 
                                                form.submission_blocks() ]
                try: block = form.blocks.get(name=self.instance.name)
                except FormBlock.DoesNotExist: return
                if block.block_type() != 'stock' or not block.stock.composite:
                    self.fields['field'].widget = forms.HiddenInput()
                    return
                
                self.fields['field'].choices = block.stock.render_choices()
            else:
                choices = []
                for block in presentation.form.collections():
                    for f in block.collection_fields():
                        if (f, f) not in choices: choices.append((f, f))
                self.fields['name'].choices = [('', '-')] + choices
                
                self.fields['field'].widget = forms.HiddenInput()


class ReferencesFormSet(forms.models.BaseInlineFormSet):
    def __init__(self, presentation=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.presentation = presentation
    
    def get_form_kwargs(self, index):
        kwargs = super().get_form_kwargs(index)
        kwargs['presentation'] = self.presentation
        return kwargs
        
    @property
    def empty_form(self):
        form = self.form(
            auto_id=self.auto_id,
            prefix=self.add_prefix('__prefix__'),
            empty_permitted=True,
            use_required_attribute=False,
            **self.get_form_kwargs(None),
            renderer=self.renderer,
            extra=True
        )
        self.add_fields(form, None)
        return form


class PresentationForm(AdminJSONForm):
    custom_css = forms.CharField(required=False, max_length=5000,
                                 widget=widgets.AdminTextareaWidget(attrs={
        'style': 'font-family: monospace;'
    }))
    hide_stats = NegatedBooleanField(
        label='show progress stats', required=False
    )
    
    class Meta:
        json_fields = {'options': ['custom_css', 'hide_stats']}


class MetricForm(forms.ModelForm):
    class Meta:
        help_texts = {
            'display_values': 'For panelist view or PDF export, list values ' \
                              'instead of count.'
        }
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        if 'instance' in kwargs and kwargs['instance']:
            if (
                kwargs['instance'].type != Metric.MetricType.COUNT
                or kwargs['instance'].count_value is not None
            ):
                self.fields['display_values'].widget = forms.HiddenInput()


class CohortForm(forms.ModelForm):
    def clean(self):
        cleaned_data = super().clean()
        status = cleaned_data['status']
        if status == Cohort.Status.INACTIVE: return cleaned_data
        
        if not cleaned_data['panel']:
            self.add_error('panel',
                           'Must be specified, unless cohort is inactive.')
        if status == Cohort.Status.ACTIVE and not cleaned_data['presentation']:
            self.add_error('presentation',
                           'Must be specified for an active cohort.')
        if status == Cohort.Status.ACTIVE:
            if not cleaned_data['inputs']:
                self.add_error('inputs', 'There has to be a primary input.')
            else:
                inputs = cleaned_data['inputs']
                qs = Input.objects.filter(pk__in=inputs).order_by('_rank')
                primary = qs[:1]
                if not primary or primary[0].type == Input.InputType.TEXT:
                    self.add_error('inputs',
                                   "Primary input can't be a text input.")
        return cleaned_data


class MetricsExportForm(ExportAdminForm):
    def __init__(self, metrics=None, inputs=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        for metric in metrics or []:
            name = f'metric_{metric.input.name}_{metric.name}'
            label = f'{metric.input.name} {metric.name}'
            self.fields[name] = forms.BooleanField(required=False, label=label,
                                                   initial=True)
        for input in inputs or []:
            name = 'input_' + input.name
            self.fields[name] = forms.ChoiceField(label=input.name,
                choices=[('no', 'not included'),
                         ('combine', 'in one column')]
            )


CC_CHOICES = [('no', 'not included'), ('repeat', 'repeated columns'),
               ('combine', 'in one column')]

class CombinedExportForm(forms.Form):
    fixed_collections = forms.ChoiceField(choices=CC_CHOICES, initial='combine')
    file_collections = forms.ChoiceField(choices=CC_CHOICES)
    nonfile_collections = forms.ChoiceField(choices=CC_CHOICES)
    metrics = forms.BooleanField(required=False, initial=True)
    text_inputs = forms.BooleanField(required=False, label='text inputs ' \
                                                           '(combined)')


class PresentationExportForm(forms.Form):
    presentation = forms.ModelChoiceField(Presentation.objects.all())
    orientation = forms.ChoiceField(choices=[ (n, n) for n in ('portrait',
                                                               'landscape') ])
    
    def __init__(self, program_form=None, metrics=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        pres = self.fields['presentation']
        if program_form: pres.queryset = pres.queryset.filter(form=program_form)
        
        for metric in metrics or []:
            name = f'metric_{metric.input.name}_{metric.name}'
            label = f'{metric.input.name} {metric.name}'
            self.fields[name] = forms.BooleanField(required=False, label=label,
                                                   initial=True)


class ScoresForm(forms.Form):
    def __init__(self, inputs=None, allow_skip=False, *args, **kwargs):
        self.inputs, self.allow_skip = inputs, allow_skip
        super().__init__(*args, **kwargs)
        
        for i, input in enumerate(inputs):
            if input.type == Input.InputType.NUMERIC:
                field = forms.IntegerField(min_value=input.min_num,
                                           max_value=input.max_num)
                if allow_skip or i: field.required = False
            elif input.type == Input.InputType.BOOLEAN:
                field = forms.BooleanField(required=False)
            else:
                field = forms.CharField(max_length=input.max_chars)
                field.widget.attrs['placeholder'] = input.label
                if not input.min_num: field.required = False
            
            field.label = input.label
            if not i: field.widget.attrs['class'] = 'primary-input'
            self.fields[input.name] = field
    
    def clean(self):
        cleaned_data = super().clean()
        if not self.allow_skip: return cleaned_data
        
        for input in self.inputs:
            val = cleaned_data.get(input.name, None)
            if not val and input.type != Input.InputType.TEXT:
                cleaned_data[input.name] = 0 # zero for skipped
        
        return cleaned_data



from django.http import HttpResponseRedirect, HttpResponseBadRequest
from django.views import generic
from django.core.paginator import Paginator
from django.db.models import F, Q, Count, Exists, Subquery, OuterRef
from django.db.models.functions import Coalesce
from django.db.models.lookups import Exact
from django.urls import reverse
from django.shortcuts import get_object_or_404
from django.contrib.auth.mixins import LoginRequiredMixin, UserPassesTestMixin
from django.contrib.contenttypes.models import ContentType
from random import random

from formative.models import Program, Form
from .forms import ScoresForm
from .models import Cohort, CohortMember, Score, Input, Presentation


URL_PREFIX = 'plugins:reviewpanel:'
SCORES_PER_PAGE = 50


class ProgramView(LoginRequiredMixin, generic.DetailView):
    model = Program
    context_object_name = 'program'
    template_name = 'reviewpanel/program_index.html'
    show_all = False
    
    def get_object(self):
        if self.show_all: return None
        return super().get_object()
    
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        
        if self.show_all:
            programs = Program.objects.filter(
                form__cohort__panel__panelists=self.request.user
            ).exclude(form__cohort__status=Cohort.Status.INACTIVE)
        else: programs = [self.object]
        
        through, user = Cohort.inputs.through, self.request.user
        input_q = through.objects.filter(cohort=OuterRef('cohort'))
        primary = Subquery(input_q.order_by('input___rank').values('input')[:1])
        
        forms = {}
        for program in programs:
            q = Score.objects.filter(panelist=user, cohort=OuterRef('pk'))
            scored = q.annotate(p=primary).filter(input=F('p'))
            scored = scored.filter(Q(value__gt=0) |
                                   Q(input__type=Input.InputType.BOOLEAN))
            cohort_scored = scored.exclude(value=None).values('cohort')
            scored_count = cohort_scored.annotate(c=Count('*')).values('c')
            cohorts_qs = Cohort.objects.filter(form__program=program,
                                               panel__panelists=user)
            vals = cohorts_qs.values('form__slug', 'form__name', 'status')
            cohort_counts = vals.annotate(count=Coalesce(Subquery(scored_count),
                                                         0))
            forms[program] = {}
            for c in cohort_counts:
                default = {'active_scored': 0, 'completed_scored': 0,
                           'active': False}
                form = forms[program].setdefault(c['form__slug'], default)
                form['name'] = c['form__name']
                if c['status'] == Cohort.Status.ACTIVE:
                    form['active_scored'] = c['count']
                    form['active'] = True
                elif c['status'] == Cohort.Status.COMPLETED:
                    form['completed_scored'] = c['count']
        
        context['forms'] = forms
        return context


class FormObjectMixin(generic.detail.SingleObjectMixin):
    context_object_name = 'program_form'
    slug_url_kwarg = 'form_slug'
    
    def get_queryset(self):
        return Form.objects.filter(program__slug=self.kwargs['program_slug'])


class FormInfoView(LoginRequiredMixin, FormObjectMixin, generic.DetailView):
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        
        cohorts = Cohort.objects.filter(panel__panelists=self.request.user,
                                        form=self.object)
        cohorts = cohorts.exclude(status=Cohort.Status.INACTIVE)
        messages, cohort_active = ([], []), False
        for cohort in cohorts:
            active = int(cohort.status == Cohort.Status.ACTIVE)
            if active: cohort_active = True
            if cohort.message not in messages[active]:
                messages[active].append(cohort.message)
        messages = messages[int(cohort_active)]
        if len(messages) == 1 and not messages[0]: messages[0] = 'All done!'
        
        context.update(cohort_active=cohort_active, messages=messages)
        if 'closed' in self.template_name: return context
        
        if not cohort_active or 'completed' in self.request.GET:
            cohorts = cohorts.filter(status=Cohort.Status.COMPLETED)
            context['completed'] = True
        else: cohorts = cohorts.filter(status=Cohort.Status.ACTIVE)
        
        through = Cohort.inputs.through
        input_q = through.objects.filter(cohort=OuterRef('cohort'))
        primary = Subquery(input_q.order_by('input___rank').values('input')[:1])
        user_scores = Score.objects.filter(panelist=self.request.user,
                                           cohort__in=cohorts.values('pk'),
                                           form=self.object).exclude(value=None)
        scores = user_scores.annotate(pri=primary).filter(input=F('pri'))
        scores = scores.select_related('input')
        
        paginator = Paginator(scores.order_by('created'), SCORES_PER_PAGE)
        page = self.request.GET.get('page', 1)
        context['page'] = paginator.get_page(page)
        return context


class FormView(LoginRequiredMixin, generic.RedirectView, FormObjectMixin):
    permanent = False
    
    def panelist_cohorts(self, user):
        # find the cohorts with unseen apps reviewed by panels this user is on
        cohorts = Cohort.objects.filter(status=Cohort.Status.ACTIVE,
                                        form=self.object, panel__panelists=user)
        
        members_subq = CohortMember.objects.filter(cohort=OuterRef('pk'))
        
        # subquery to find a unique cohort for each applicant (overlap possible)
        q = cohorts.filter(cohortmember__object_id=OuterRef('object_id'))
        app_cohort = Subquery(q.order_by('size', '-created').values('pk')[:1])
        apps = members_subq.annotate(app_cohort=app_cohort)
        
        # annotate cohorts with whether there's app w/ no input for that cohort
        app_cohort_score_subsubq = Score.objects.exclude(value=None).filter(
            panelist=user, content_type=OuterRef('content_type'),
            object_id=OuterRef('object_id'), cohort=OuterRef('app_cohort')
        )
        unscored_subq = apps.exclude(Exists(app_cohort_score_subsubq))
        return cohorts, cohorts.filter(Exists(unscored_subq))
    
    def choose_panel(self, user, cohorts):
        total, weights = 0, {}
        for cohort in cohorts.annotate(weight=F('panel_weight') * F('size')):
            weights[cohort] = cohort.weight
            total += cohort.weight
        if not weights: return None
        if not total:
            weights, total = { cohort: 1 for cohort in weights }, len(weights)
        
        r, v = random(), 0
        for cohort, weight in weights.items():
            v += weight / total
            if r < v: return cohort
        return cohort
    
    def get_redirect_url(self, *args, **kwargs):
        self.object = self.get_object()
        form, user = self.object, self.request.user
        
        active_cohorts, cohorts = self.panelist_cohorts(user)
        
        cohort = self.choose_panel(user, cohorts)
        if not cohort:
            scores = Score.objects.exclude(value=None)
            skipped = scores.filter(panelist=user, value=0,
                                    cohort__in=active_cohorts).order_by('?')
            skipped = skipped.exclude(input__type=Input.InputType.BOOLEAN)[:1]
            if not skipped:
                return reverse(URL_PREFIX + 'form_complete', kwargs=kwargs)
            kwargs['pk'] = str(skipped[0].object_id)
            return reverse(URL_PREFIX + 'submission_skips', kwargs=kwargs)
        
        unscored = None
        try: unscored = Score.objects.get(panelist=user, value=None)
        except Score.DoesNotExist: pass
        
        if unscored: cohort, unscored_id = unscored.cohort, unscored.object_id
        else: unscored_id = None
        
        # TODO: seen on primary input, only
        user_seen = Score.objects.exclude(value=None).filter(panelist=user)
        active_user_seen = user_seen.filter(cohort__status=Cohort.Status.ACTIVE)
        members_seen = active_user_seen.values('object_id')
        
        input = cohort.inputs.order_by('_rank')[0]
        scores = Score.objects.filter(value__gt=0,
                                      form=cohort.form, input=input)
        app_scores = scores.filter(object_id=OuterRef('object_id'))
        app_counts = app_scores.values('object_id').annotate(count=Count('*'))
        counts, members = app_counts.values('count'), cohort.cohortmember_set
        
        q = cohorts.filter(cohortmember__object_id=OuterRef('object_id'))
        app_cohort = Subquery(q.order_by('size', '-created').values('pk')[:1])
        app_members = members.annotate(app_cohort=app_cohort)
        cohort_apps = app_members.filter(app_cohort=cohort.pk)
        
        not_seen = cohort_apps.exclude(object_id__in=Subquery(members_seen))
        apps = not_seen.annotate(scores=Coalesce(Subquery(counts), 0),
                                 put_first=Exact(F('object_id'), unscored_id))
        chosen, first = None, None
        for member in apps.order_by('-put_first', 'scores', '?')[:2]:
            if member.put_first:
                first = member
                continue
            lowest_count = member.scores
            if first and unscored:
                if first.scores <= lowest_count: chosen = first
                else: unscored.delete() # panelist waited; it will come up later
            if not chosen: chosen = member
        if not chosen: chosen = first
        
        score = Score.objects.create_for_cohort(user, cohort, input=input,
                                                submission=chosen.member)
        if chosen != first: score.save()
        kwargs['pk'] = str(chosen.object_id)
        return reverse('plugins:reviewpanel:submission', kwargs=kwargs)


class SubmissionObjectMixin(generic.detail.SingleObjectMixin):
    context_object_name = 'submission'
    
    def get_queryset(self):
        form = get_object_or_404(Form.objects.select_related('program'),
                                 program__slug=self.kwargs['program_slug'],
                                 slug=self.kwargs['form_slug'])
        
        form.model, form.item_model
        return form.model.objects.all()


class PresentationContextMixin:
    def presentation_context(self, pres):
        form, references = pres.form, pres.references.order_by('_rank')
        names = Subquery(references.filter(collection='').values('name'))
        cnames = Subquery(references.exclude(collection='').values('collection'))
        blocks = form.blocks.filter(Q(name__in=names) | Q(name__in=cnames))
        
        items = {}
        if self.object and form.item_model:
            citems = self.object._items.filter(_collection__in=cnames)
            items = self.object._collections(queryset=citems, form=form)
        
        refs, select_refs = {}, {}
        for ref in references:
            refs.setdefault(ref.section.name, []).append(ref)
            
            if not ref.collection: continue
            if ref.collection not in items: items[ref.collection] = []
            section = ref.select_section if ref.select_section else ref.section
            section_refs = select_refs.setdefault(section.name, {})
            collection = section_refs.setdefault(ref.collection, ([], []))
            
            if ref.is_file: collection[0].append(ref)
            else: collection[1].append(ref)
        
        sections = []
        for section in pres.template.sections.order_by('-y'):
            selectors = {}
            # a map from collection to array of refs that want a selector here
            if section.name in select_refs:
                selectors = { col: (v[0]+v[1], bool(v[0])) # + whether have file
                              for col, v in select_refs[section.name].items() }
            sections.append((section,
                             refs[section.name] if section.name in refs else [],
                             selectors))
        return {
            'blocks': { b.name: b for b in blocks },
            'items': items, 'sections': sections
        }
            
    
class SubmissionDetailView(LoginRequiredMixin, SubmissionObjectMixin,
                           generic.DetailView, PresentationContextMixin):
    template_name = 'reviewpanel/submission.html'
    
    def render_to_response(self, context, **kwargs):
        user = self.request.user
        entries = CohortMember.objects.filter(object_id=self.object.pk,
                                              cohort__panel__panelists=user)
        if not entries.exclude(cohort__status=Cohort.Status.INACTIVE).exists():
            return self.handle_no_permission()
        
        if 'cohort' not in context:
            kwargs = { k: self.kwargs[k] for k in self.kwargs if k != 'pk' }
            return HttpResponseRedirect(reverse('plugins:reviewpanel:form',
                                                kwargs=kwargs))
        
        return super().render_to_response(context, **kwargs)
    
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        score, prev, user = None, None, self.request.user
        prev_on, next_on = False, False
        
        scores = Score.objects.filter(object_id=self.object.pk, panelist=user)
        try: score = scores.get(value=None) # placeholder for an assigned member
        except Score.DoesNotExist: pass
        
        query = scores.filter(cohort__status=Cohort.Status.ACTIVE)
        try: prev = query.exclude(value=None).order_by('-created')[:1].get()
        except Score.DoesNotExist: pass

        if score and score.cohort.status != Cohort.Status.ACTIVE:
            score.delete()
            return context
        if prev and score: # cohort scored w before, made active again
            score.delete() # so delete the assignment and show it as prev scored
        if prev:
            score = prev
            qs = Score.objects.filter(panelist=user, form=score.form)
            prior = qs.filter(Q(created__lt=score.created) |
                             Q(created=score.created) & Q(id__lt=score.id))
            prev_on, next_on = prior.exists(), True
        if not score: return context # TODO: allow link sharing option
        
        query = Cohort.objects.select_related('presentation__template')
        cohort = query.get(pk=score.cohort_id)
        form, inputs = cohort.presentation.form, cohort.inputs.order_by('_rank')
        
        apps = CohortMember.objects.filter(cohort__panel__panelists=user,
                                           cohort__status=Cohort.Status.ACTIVE,
                                           cohort__form=form)
        apps_count = apps.values('object_id').distinct().count()
        
        through = Cohort.inputs.through
        input_q = through.objects.filter(cohort=OuterRef('cohort'))
        primary = Subquery(input_q.order_by('input___rank').values('input')[:1])
        qs = Score.objects.filter(value__isnull=False, panelist=user, form=form,
                                  cohort__panel__panelists=user,
                                  cohort__status=Cohort.Status.ACTIVE)
        scored = qs.annotate(pri=primary,
                             skip=Exact(F('value'), 0)).filter(input=F('pri'))
        scored_counts = scored.values('skip').annotate(c=Coalesce(Count('*'),
                                                                  0))
        counts = {False: 0, True: 0}
        for c in scored_counts.values('skip', 'c'): counts[c['skip']] += c['c']
        
        initial = {}
        if score.value is not None:
            for score in scores.select_related('input').filter(cohort=cohort):
                val = score.value
                if score.input.type == Input.InputType.TEXT: val = score.text
                if val: initial[score.input.name] = val
        scores_form = ScoresForm(inputs=inputs, allow_skip=cohort.allow_skip,
                                 initial=initial)
        
        context.update(self.presentation_context(cohort.presentation))
        context.update({
            'cohort': cohort, 'presentation': cohort.presentation,
            'template': cohort.presentation.template, 'form': scores_form,
            'prev_on': prev_on, 'next_on': next_on,
            'stats': {'scored': counts[False], 'skipped': counts[True],
                      'total': apps_count}
        })
        return context


class ScoresFormView(LoginRequiredMixin, SubmissionObjectMixin,
                     generic.FormView):
    template_name = 'reviewpanel/submission.html'
    form_class = ScoresForm
    skips = False
    
    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        
        self.inputs = self.cohort.inputs.order_by('_rank')
        kwargs.update(inputs=self.inputs, allow_skip=self.cohort.allow_skip)
        return kwargs
    
    def form_invalid(self, form):
        # shouldn't be possible unless there's form tampering
        return HttpResponseRedirect(request.get_full_path())
    
    def navigate_history(self, queryset, score, prev=False):
        qs = queryset
        if prev: qs = qs.filter(Q(created__lt=score.created) |
                                Q(created=score.created) & Q(id__lt=score.id))
        else: qs = qs.filter(Q(created__gt=score.created) |
                             Q(created=score.created) & Q(id__gt=score.id))
        try: return qs[0]
        except IndexError: return None
    
    def form_valid(self, form):
        user, program_form = self.request.user, self.cohort.form
        
        score = None
        for i, input in enumerate(self.inputs):
            ctype = ContentType.objects.get_for_model(program_form.model)
            args = {'content_type': ctype, 'form': program_form}
            python_val, args['text'] = form.cleaned_data[input.name], ''
            if input.type == Input.InputType.TEXT:
                args['value'], args['text'] = int(bool(python_val)), python_val
            else: args['value'] = int(python_val)
            
            kwargs = {'panelist': user, 'object_id': self.submission.pk,
                      'cohort': self.cohort, 'input': input}
            if not i:
                program_form, id = self.cohort.form, self.submission.pk
                try: score = Score.objects.get(**kwargs)
                except Score.DoesNotExist: return HttpResponseBadRequest()
                
                score.value, score.text = args['value'], args['text']
                score.save()
            elif input.type == Input.InputType.TEXT and not python_val:
                Score.objects.filter(**kwargs).delete()
            else: Score.objects.update_or_create(defaults=args, **kwargs)
        
        request = self.request
        nav = 'prev_scored' in request.POST or 'next_scored' in request.POST
        if nav and not self.skips:
            qs = Score.objects.filter(panelist=user, form=program_form)
            previous = 'prev_scored' in request.POST
            target = self.navigate_history(qs, score, prev=previous)
            
            if not target:
                self.kwargs.pop('pk')
                url = reverse(URL_PREFIX + 'form_index', kwargs=self.kwargs)
            else:
                self.kwargs['pk'] = target.object_id
                url = reverse(URL_PREFIX + 'submission', kwargs=self.kwargs)
            return HttpResponseRedirect(url)
        
        # get another random submission to review
        self.kwargs.pop('pk')
        return FormView.as_view()(self.request, **self.kwargs)
    
    def post(self, request, *args, **kwargs):
        self.submission = self.get_object()
        
        if 'cohort_id' not in request.POST: return HttpResponseBadRequest()
        
        try: self.cohort = Cohort.objects.get(pk=int(request.POST['cohort_id']),
                                              panel__panelists=request.user)
        except ValueError: return HttpResponseBadRequest()
        if self.cohort.status != Cohort.Status.ACTIVE: # this will redir to an
            self.kwargs.pop('pk') # app in cohort still active, or form_complete
            return FormView.as_view()(self.request, **self.kwargs)
        
        return super().post(request, *args, **kwargs)


class SubmissionView(generic.View):
    skips = False
    
    def get(self, request, *args, **kwargs):
        return SubmissionDetailView.as_view()(request, *args, **kwargs)
    
    def post(self, req, *args, **kwargs):
        return ScoresFormView.as_view(skips=self.skips)(req, *args, **kwargs)


class SubmissionAdminView(UserPassesTestMixin, SubmissionObjectMixin,
                          generic.DetailView, PresentationContextMixin):
    template_name = 'reviewpanel/submission.html'
    
    def test_func(self):
        return self.request.user.is_staff
    
    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        
        presentation = get_object_or_404(Presentation,
                                         pk=self.kwargs['presentation'])
        
        context.update(self.presentation_context(presentation))
        context.update({
            'presentation': presentation, 'template': presentation.template
        })
        return context


class PresentationAdminView(SubmissionAdminView):
    def get_queryset(self):
        form = get_object_or_404(Form.objects.select_related('program'),
                                 program__slug=self.kwargs['program_slug'],
                                 slug=self.kwargs['form_slug'])
        
        form.model, form.item_model
        return form.model.objects.filter(_submitted__isnull=False).order_by('?')
    
    def get_object(self):
        queryset = self.get_queryset()
        if not queryset: return None
        return queryset[0]



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _

from . import __version__


class ReviewPanelConfig(AppConfig):
    name = 'reviewpanel'
    verbose_name = 'reviewpanel'
    
    class FormativePluginMeta:
        name = 'reviewpanel'
        author = 'John Kyle Cronan'
        description = _('Review and score application form submissions')
        version = __version__
    
    def ready(self):
        from . import signals
    
    def is_available(self, form=None):
        return False



# Generated by Django 4.0.3 on 2022-03-26 16:25

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('reviewpanel', '0002_alter_input_options_cohort_allow_skip'),
    ]

    operations = [
        migrations.AlterField(
            model_name='cohort',
            name='inputs',
            field=models.ManyToManyField(blank=True, related_name='cohorts', related_query_name='cohort', to='reviewpanel.input'),
        ),
        migrations.AlterField(
            model_name='panel',
            name='panelists',
            field=models.ManyToManyField(blank=True, related_name='panels', related_query_name='panel', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='templatesection',
            name='scroll',
            field=models.BooleanField(default=False),
        ),
    ]



from django.dispatch import receiver
from django.contrib.contenttypes.models import ContentType

from formative.admin import site
from formative.models import Program, Form
from formative.signals import form_published_changed, register_user_actions, \
    all_submissions_pre_delete, all_forms_unpublish
from .admin import add_to_panel, ProgramFormsAdmin, FormSubmissionsAdmin
from .models import CohortMember, Score


programs_registered, forms_registered = {}, {}

def create_proxy_model(base, name, **kwargs):
    class Meta:
        proxy = True
        app_label = 'reviewpanel'
    
    model = type(name, (base,), {'__module__': 'reviewpanel', 'Meta': Meta})
    for k, v in kwargs.items(): setattr(model._meta, k, v)
    return model

@receiver(form_published_changed, dispatch_uid='reviewpanel_published_changed')
def form_published_changed(sender, **kwargs):
    for model in programs_registered: site.unregister(model)
    programs_registered.clear()
    for model in forms_registered: site.unregister(model)
    forms_registered.clear()
    
    queryset = Form.objects.exclude(status=Form.Status.DRAFT)
    for form in queryset.select_related('program'):
        slug = form.program.slug
        if slug not in [ p._meta.program_slug for p in programs_registered ]:
            model = create_proxy_model(Form, form.program.db_slug + '_forms',
                                       program_slug=slug,
                                       verbose_name=slug+' form',
                                       verbose_name_plural=slug+' forms')
            site.register(model, ProgramFormsAdmin)
            programs_registered[model] = True
        
        model = create_proxy_model(form.model,
                                   form.program.db_slug + '_' + form.db_slug,
                                   program_slug=slug, form_slug=form.slug,
                                   verbose_name=form.slug+' submission',
                                   verbose_name_plural=form.slug+' submissions')
        site.register(model, FormSubmissionsAdmin)
        forms_registered[model] = True

@receiver(register_user_actions, dispatch_uid='reviewpanel_user_action')
def register_user_actions(sender, **kwargs):
    return {'add_to_panel': add_to_panel}

@receiver(all_submissions_pre_delete, dispatch_uid='reviewpanel_pre_delete')
def all_submissions_pre_delete(sender, instance, **kwargs):
    match = None
    for model in forms_registered:
        if model._meta.db_table == sender._meta.db_table: match = True
    if not match: return # all_forms_unpublish has already taken care of it
    
    CohortMember.objects.filter(object_id=instance.pk).delete()
    Score.objects.filter(object_id=instance.pk).delete()

@receiver(all_forms_unpublish, dispatch_uid='reviewpanel_form_unpublish')
def all_forms_unpublish(sender, content_type, **kwargs):
    form = sender
    CohortMember.objects.filter(content_type=content_type).delete()
    Score.objects.filter(content_type=content_type).delete()
    
    # TODO: deactivate any active cohorts



from django import forms
from django.contrib import admin, messages
from django.contrib.admin.views.main import ChangeList
from django.contrib.contenttypes.models import ContentType
from django.db.models import F, Q, Count, Exists, Subquery, OuterRef
from django.db.models.functions import Coalesce
from django.http import HttpResponseRedirect
from django.template.response import TemplateResponse
from django.shortcuts import get_object_or_404
from django.urls import path, reverse
from django.utils.safestring import mark_safe
from django_admin_inline_paginator.admin import TabularInlinePaginated
from functools import partial
import types

from formative.admin import site
from formative.models import Form, SubmissionRecord
from .forms import ReferencesFormSet, ReferenceForm, MetricForm, CohortForm, \
    PresentationForm, MetricsExportForm, CombinedExportForm, \
    PresentationExportForm
from .models import Template, TemplateSection, Reference, Presentation, Input, \
    Panel, Cohort, CohortMember, Score, Metric
from .utils import MetricsTabularExport, CombinedTabularExport, \
     PresentationPrintExport


class TemplateSectionInline(admin.StackedInline):
    model = TemplateSection
    extra = 0


@admin.register(Template, site=site)
class TemplateAdmin(admin.ModelAdmin):
    list_display = ('name', 'program')
    list_filter = ('program',)
    inlines = [TemplateSectionInline]


class ReferenceInline(admin.StackedInline):
    model = Reference
    extra = 0
    formset = ReferencesFormSet
    form = ReferenceForm
    
    def get_formset(self, request, obj=None, **kwargs):
        request._obj_ = obj # TODO the code below could be moved into Form class
        kwargs['formfield_callback'] = partial(self.formfield_for_dbfield,
                                               request=request)
        return super().get_formset(request, obj, **kwargs)
    
    def formfield_for_dbfield(self, db_field, **kwargs):
        field = super().formfield_for_dbfield(db_field, **kwargs)
        pres = kwargs.pop('request')._obj_
        
        if db_field.name in ('section', 'select_section'):
            if pres.template:
                field.queryset = field.queryset.filter(template=pres.template)
            else: field.queryset = field.queryset.none()
        
        elif db_field.name == 'collection':
            choices = [('', '-')]
            for b in pres.form.collections():
                if (b.name, b.name) not in choices:
                    choices.append((b.name, b.name))
            return forms.ChoiceField(choices=choices, required=False, **kwargs)
        
        elif db_field.name == 'name':
            regular_blocks = pres.form.submission_blocks()
            choices = [('', '-')] + [ (n, n) for n in regular_blocks ]
            for b in pres.form.collections():
                for f in b.collection_fields():
                    if (f, f) not in choices: choices.append((f, f))
            return forms.ChoiceField(choices=choices, required=False, **kwargs)
        
        elif db_field.name == 'field':
            return forms.ChoiceField(choices=(), required=False, **kwargs)
        
        return field


@admin.register(Presentation, site=site)
class PresentationAdmin(admin.ModelAdmin):
    form = PresentationForm
    list_display = ('name', 'created', 'template', 'form')
    list_filter = ('form',)
    inlines = [ReferenceInline]
    
    def get_formset_kwargs(self, request, obj=None, *args):
        kwargs = super().get_formset_kwargs(request, obj, *args)
        kwargs['presentation'] = obj
        return kwargs
    
    def get_formsets_with_inlines(self, request, obj=None):
        for inline in self.get_inline_instances(request, obj):
            if not isinstance(inline, ReferenceInline) or obj is not None:
                yield inline.get_formset(request, obj), inline
    
    def get_deleted_objects(self, objs, request):
        to_del, models, perms, protected = super().get_deleted_objects(objs,
                                                                       request)
        for obj in objs:
            for cohort in obj.cohorts.filter(status=Cohort.Status.ACTIVE):
                protected.append(cohort)
        return to_del, models, perms, protected
    
    def view_on_site(self, obj):      # don't need the sites framework
        return obj.get_absolute_url() # getting in the way at the moment


class MetricInline(admin.StackedInline):
    model = Metric
    form = MetricForm
    extra = 0


@admin.register(Input, site=site)
class InputAdmin(admin.ModelAdmin):
    list_display = ('name', 'form')
    list_filter = ('form',)
    exclude = ('_rank',)
    inlines = [MetricInline]


@admin.action(description='Add users to panel')
def add_to_panel(modeladmin, request, queryset):
    if 'panel' in request.POST:
        panel = get_object_or_404(Panel, id=int(request.POST['panel']))
        panel.panelists.add(*queryset)
        msg = f'Users added to panel "{panel.name}".',
        modeladmin.message_user(request, msg, messages.SUCCESS)
        return HttpResponseRedirect(request.get_full_path())
    
    class PanelForm(forms.Form):
        panel = forms.ModelChoiceField(queryset=Panel.objects.all())
    
    template_name = 'admin/reviewpanel/add_to_panel.html'
    context = {
        **modeladmin.admin_site.each_context(request),
        'opts': modeladmin.model._meta, 'users': queryset,
        'form': PanelForm(), 'title': 'Add to panel'
    }
    return TemplateResponse(request, template_name, context)


class PanelistInline(admin.TabularInline):
    model = Panel.panelists.through
    extra = 0
    verbose_name = 'panelist'
    verbose_name_plural = 'panelists'


@admin.register(Panel, site=site)
class PanelAdmin(admin.ModelAdmin):
    list_display = ('name', 'program')
    list_filter = ('program',)
    inlines = [PanelistInline]
    exclude = ('panelists',)
    
    def get_deleted_objects(self, objs, request):
        to_del, models, perms, protected = super().get_deleted_objects(objs,
                                                                       request)
        for obj in objs:
            for cohort in obj.cohorts.exclude(status=Cohort.Status.INACTIVE):
                protected.append(cohort)
        return to_del, models, perms, protected


class CohortMemberInline(TabularInlinePaginated):
    model = CohortMember
    extra = 0
    per_page = 100
    can_delete = True
    exclude = ('content_type', 'object_id')
    readonly_fields = ('email', 'submitted')
    
    def has_add_permission(self, request, obj=None):
        return False
    
    def get_queryset(self, request):
        queryset = super().get_queryset(request)
        match = request.resolver_match
        cohort = Cohort.objects.get(pk=match.kwargs['object_id'])
        model = cohort.form.model
        if not model: return queryset
        
        obj = model.objects.values('pk').filter(pk=OuterRef('object_id'))
        qs = queryset.annotate(email=Subquery(obj.values('_email')),
                               submitted=Subquery(obj.values('_submitted')))
        return qs.order_by('-submitted')
    
    def email(self, object):
        return object.email # TODO: link to the submission change view
    
    def submitted(self, object):
        return object.submitted


@admin.register(Cohort, site=site)
class CohortAdmin(admin.ModelAdmin):
    form = CohortForm
    list_display = ('name', 'panel', 'status', 'form')
    list_filter = ('panel', 'status', 'form')
    inlines = [CohortMemberInline]
    readonly_fields = ('primary_input', 'size')
    
    def get_formsets_with_inlines(self, request, obj=None):
        for inline in self.get_inline_instances(request, obj):
            if not isinstance(inline, CohortMemberInline) or obj is not None:
                yield inline.get_formset(request, obj), inline
    
    def save_related(self, request, form, formsets, change):
        super().save_related(request, form, formsets, change)
        if change: # save the cohort again with an updated size count
            obj = form.instance
            obj.size = obj.cohortmember_set.count()
            self.save_model(request, obj, form, change)
    
    def primary_input(self, obj):
        input = obj.inputs.order_by('_rank')[:1]
        return input[0] if input else None


class ScoreTypeFilter(admin.SimpleListFilter):
    title = 'value'
    parameter_name = 'value'
    
    def lookups(self, request, model_admin):
        return (('yes', 'scored'), ('skip', 'skipped'), ('no', 'unscored'))
    
    def queryset(self, request, queryset):
        bool_type = Input.InputType.BOOLEAN
        if self.value() == 'yes':
            return queryset.filter(Q(value__gt=0) |
                                   Q(value=0) & Q(input__type=bool_type))
        elif self.value() == 'skip':
            return queryset.filter(value=0).exclude(input__type=bool_type)
        elif self.value() == 'no': return queryset.filter(value__isnull=True)


@admin.register(Score, site=site)
class ScoreAdmin(admin.ModelAdmin):
    list_display = ('submission', 'panelist', 'input', 'cohort', 'display_val',
                    'created')
    list_filter = ('panelist', 'input', 'cohort', 'form', ScoreTypeFilter)
    readonly_fields = ('submission_link',)
    
    def get_fields(self, request, obj=None):
        fields = super().get_fields(request, obj)
        if obj: return tuple(f for f in fields
                             if f not in ('content_type', 'object_id'))
        return fields
    
    def get_readonly_fields(self, request, obj=None):
        fields = super().get_readonly_fields(request, obj)
        if obj: fields += ('panelist', 'form')
        return fields
    
    @admin.display(ordering='value', description='value')
    def display_val(self, obj):
        if obj.input.type == Input.InputType.TEXT: return obj.text
        if obj.value is None: return '[unscored]'
        if obj.input.type == Input.InputType.BOOLEAN: return bool(obj.value)
        if not obj.value: return '[skipped]'
        return obj.value
    
    @admin.display(description='submission')
    def submission_link(self, obj):
        name = obj.form.program.db_slug + '_' + obj.form.db_slug
        url = reverse('admin:reviewpanel_%s_change' % (name,),
                      args=(obj.object_id,),
                      current_app=self.admin_site.name)
        return mark_safe(f'<a href="{url}">{obj.submission}</a>')


class FormChangeList(ChangeList):
    def url_for_result(self, result):
        name = result.program.db_slug + '_' + result.db_slug
        return reverse('admin:%s_%s_changelist' % (self.opts.app_label, name),
                       current_app=self.model_admin.admin_site.name)

class ProgramFormsAdmin(admin.ModelAdmin):
    list_display = ('name', 'submitted', 'created', 'modified')
    list_select_related = ('program',)
    actions = ['export_ods']
    
    def has_add_permission(self, request): return False
    def has_change_permission(self, request, obj=None): return False
    def has_delete_permission(self, request, obj=None): return False
    
    def get_changelist(self, request, **kwargs):
        return FormChangeList
    
    def get_queryset(self, request):
        slug = self.model._meta.program_slug
        forms = self.model.objects.exclude(status=Form.Status.DRAFT)
        forms = forms.filter(program__slug=slug)
        
        # w/ SubmissionRecord we can get totals for multiple forms using 1 table
        args = {'program__slug': slug, 'form': OuterRef('slug')}
        args.update(type=SubmissionRecord.RecordType.SUBMISSION, deleted=False)
        subquery = SubmissionRecord.objects.filter(**args)
        count = subquery.values('form').annotate(count=Count('*'))
        annotation = Coalesce(Subquery(count.values('count')), 0)
        return forms.annotate(submitted=annotation)
    
    def submitted(self, obj):
        return obj.submitted
    
    @admin.action(description='Export form submissions as ODS')
    def export_ods(self, request, queryset):
        if '_export' in request.POST:
            filename = f'{self.model._meta.program_slug}_export_selected.ods'
            args = { k: request.POST[k] for k in request.POST
                     if k in ('metrics',
                              'text_inputs') or k.endswith('_collections') }
            export = CombinedTabularExport(queryset, **args)
            return export.response_ods(filename, queryset)
        
        template_name = 'admin/reviewpanel/export_forms.html'
        context = {
            **self.admin_site.each_context(request),
            'opts': self.model._meta, 'media': self.media,
            'forms': queryset, 'title': 'Export Form Submissions',
            'form': CombinedExportForm()
        }
        return TemplateResponse(request, template_name, context)


class BaseScoreFormSet(forms.BaseModelFormSet):
    def __init__(self, instance=None, *args, **kwargs):
        self.instance = instance
        super().__init__(*args, **kwargs)
        scores = Score.objects.filter(object_id=instance.pk)
        self.queryset = scores.order_by('created')
    
    @classmethod
    def get_default_prefix(cls):
        opts = cls.model._meta
        return opts.app_label + '-' + opts.model_name


class SubmissionScoresInline(admin.TabularInline):
    model = Score
    formset = BaseScoreFormSet
    readonly_fields = ('display_val', 'created',)
    
    def has_add_permission(self, request, obj=None): return False
    
    def get_formset(self, request, obj=None, **kwargs):
        fields = ('panelist', 'input', 'cohort',)
        
        defaults = {
            'form': self.form, 'formset': self.formset, 'fields': fields,
            'formfield_callback': partial(self.formfield_for_dbfield,
                                          request=request),
            'extra': 0, 'can_delete': False, 'can_order': False
        }
        return forms.modelformset_factory(self.model, **defaults)
    
    @admin.display(description='value')
    def display_val(self, obj):
        if obj.input.type == Input.InputType.TEXT: return obj.text
        elif obj.input.type == Input.InputType.BOOLEAN: return bool(obj.value)
        if not obj.value: return '[skipped]'
        return obj.value


class CohortListFilter(admin.SimpleListFilter):
    title = 'cohort'
    parameter_name = 'cohort'
    
    def lookups(self, request, model_admin):
        model = model_admin.model
        program_slug, slug = model._meta.program_slug, model._meta.form_slug
        for cohort in Cohort.objects.filter(form__slug=slug,
                                            form__program__slug=program_slug):
            yield (cohort.id, cohort.name)
    
    def queryset(self, request, queryset):
        if not self.value(): return queryset
        cohort_member = CohortMember.objects.filter(object_id=OuterRef('pk'),
                                                    cohort=self.value())
        return queryset.filter(Exists(cohort_member))


class FormSubmissionsAdmin(admin.ModelAdmin):
    list_display = ('submission_id', '_created', '_submitted')
    list_filter = (CohortListFilter,)
    list_per_page = 400
    inlines = [SubmissionScoresInline]
    actions = ['add_to_cohort', 'export_csv', 'export_pdf']
    
    def has_module_permission(self, request):
        return False # it's linked to by ProgramFormsAdmin, don't show in index
    
    def has_add_permission(self, request): return False
    def has_change_permission(self, request, obj=None): return False
    def has_delete_permission(self, request, obj=None): return False
    
    def get_metrics(self, request, **kwargs):
        fs, ps = self.model._meta.form_slug, self.model._meta.program_slug
        form = Form.objects.get(slug=fs, program__slug=ps)
        cohort_id = request.GET.get('cohort')
        
        metrics = Metric.objects.filter(admin_enabled=True,
                                        input__cohort__form=form, **kwargs)
        if cohort_id and cohort_id.isdigit():
            metrics = metrics.filter(input__cohort=int(cohort_id))
        return metrics.distinct(), form
    
    def get_queryset(self, request):
        queryset = self.model.objects.exclude(_submitted__isnull=True)
        
        metrics, _ = self.get_metrics(request)
        scores = Score.objects.all()
        for metric in metrics:
            name = f'metric_{metric.input.name}_{metric.name}'
            annotation = metric.annotation(scores, object_id='pk')
            queryset = queryset.annotate(**{name: annotation})
        
        return queryset
    
    def get_list_display(self, request):
        fields = list(super().get_list_display(request))
        
        metrics, _ = self.get_metrics(request)
        divisors, metric_fields = {}, []
        for metric in metrics.order_by('input', '-count_is_divisor'):
            def field_callable(desc, field, divisor_field=None):
                @admin.display(description=desc, ordering=field)
                def callable(self, obj):
                    v = getattr(obj, field)
                    if v is None: display = '-'
                    elif type(v) not in (int, bool): display = f'{v:.3f}'
                    else: display = v
                    
                    if divisor_field:
                        fv = getattr(obj, divisor_field)
                        if fv: display = f'{display} ({v/fv*100:.1f}%)'
                    return display
                return callable
            
            field_name = 'metric_' + metric.input.name + '_' + metric.name
            input_id, divisor_name = metric.input_id, None
            if metric.type == Metric.MetricType.COUNT and input_id in divisors:
                divisor_name = divisors[metric.input_id]
            if metric.count_is_divisor: divisors[input_id] = field_name

            c = field_callable(metric.name, field_name, divisor_name)
            # Django will reject the ordering if value isn't a field or method:
            setattr(self, field_name, types.MethodType(c, self))
            
            rec = (field_name, metric.position)
            for i, v in enumerate(metric_fields):
                if v[1] > metric.position:
                    metric_fields.insert(i, rec)
                    rec = None
                    break
            if rec: metric_fields.append(rec)
        
        cohort_id, end_fields = request.GET.get('cohort'), []
        if cohort_id and cohort_id.isdigit():
            pres = None
            try: pres = Presentation.objects.get(cohort__pk=int(cohort_id))
            except Presentation.DoesNotExist: pass
            if pres:
                def link_callable(presentation):
                    @admin.display(description='link')
                    def callable(obj):
                        form = presentation.form
                        args = {'program_slug': form.program.slug,
                                'form_slug': form.slug,
                                'pk': obj.pk, 'presentation': presentation.pk}
                        url = reverse('plugins:reviewpanel:submission_admin',
                                      kwargs=args)
                        return mark_safe(f'<a href="{url}">view</a>')
                    return callable
                end_fields.append(link_callable(pres))
        return fields + [ v[0] for v in metric_fields ] + end_fields
    
    @admin.display(description='ID')
    def submission_id(self, obj):
        return str(obj)
    
    @admin.action(description='Add submissions to cohort')
    def add_to_cohort(self, request, queryset):
        if 'cohort' in request.POST:
            cohort = get_object_or_404(Cohort, id=int(request.POST['cohort']))
            ctype = ContentType.objects.get_for_model(cohort.form.model)
            member_subq = CohortMember.objects.filter(cohort=cohort,
                                                      object_id=OuterRef('pk'))
            members = []
            for submission in queryset.exclude(Exists(member_subq)):
                members.append(CohortMember(cohort=cohort, content_type=ctype,
                                            object_id=submission.pk))
            CohortMember.objects.bulk_create(members)
            cohort.size = cohort.cohortmember_set.count()
            cohort.save()
            
            msg = f'Submissions added to cohort "{cohort.name}".'
            self.message_user(request, msg, messages.SUCCESS)
            return HttpResponseRedirect(request.get_full_path())
        
        fs, ps = self.model._meta.form_slug, self.model._meta.program_slug
        qs = Cohort.objects.filter(form__slug=fs, form__program__slug=ps)
        class CohortForm(forms.Form):
            cohort = forms.ModelChoiceField(queryset=qs)
        
        template_name = 'admin/reviewpanel/add_to_cohort.html'
        context = {
            **self.admin_site.each_context(request),
            'opts': self.model._meta, 'submissions': queryset,
            'form': CohortForm(), 'title': 'Add to cohort'
        }
        return TemplateResponse(request, template_name, context)
    
    @admin.action(description='Export submissions as CSV')
    def export_csv(self, request, queryset):
        metrics, program_form = self.get_metrics(request)
        if '_export' in request.POST:
            args = { k: request.POST[k] for k in request.POST
                     if k.startswith('block_') or k.startswith('collection_')
                        or k.startswith('cfield_') or k.startswith('metric_')
                        or k.startswith('input_') }
            export = MetricsTabularExport(program_form, queryset, **args)
            filename = f'{program_form.slug}_export_selected.csv'
            return export.csv_response(filename, queryset)
        
        inputs = Input.objects.filter(cohort__form=program_form,
                                      type=Input.InputType.TEXT)
        
        template_name = 'admin/formative/export_submissions.html'
        context = {
            **self.admin_site.each_context(request),
            'opts': self.model._meta, 'media': self.media,
            'submissions': queryset, 'title': 'Export Submissions Data',
            'form': MetricsExportForm(program_form=program_form,
                                      metrics=metrics, inputs=inputs)
        }
        return TemplateResponse(request, template_name, context)
    
    @admin.action(description='Export submissions as PDF')
    def export_pdf(self, request, queryset):
        metrics, form = self.get_metrics(request, panelist_enabled=True)
        if '_export' in request.POST:
            pres = get_object_or_404(Presentation,
                                     id=int(request.POST['presentation']))
            
            filename = f'{form.slug}_export.pdf'
            args = { k: request.POST[k] for k in request.POST
                     if k == 'orientation' or k.startswith('metric_') }
            export = PresentationPrintExport(filename, pres, **args)
            return export.response(queryset)
        
        template_name = 'admin/reviewpanel/report_submissions.html'
        context = {
            **self.admin_site.each_context(request),
            'opts': self.model._meta, 'media': self.media,
            'submissions': queryset, 'title': 'Export Submissions Report',
            'form': PresentationExportForm(program_form=form, metrics=metrics)
        }
        return TemplateResponse(request, template_name, context)



from django import template


register = template.Library()

@register.simple_tag
def collection_items(items, collection):
    return items[collection]

@register.simple_tag(takes_context=True)
def dereference_block(context, ref, submission):
    block_name = ref.name if not ref.collection else ref.collection
    if not block_name: return ''
    block = context['blocks'][block_name]
    
    if block.block_type() == 'custom':
        return getattr(submission, block_name) or ''
    elif block.block_type() == 'stock':
        stock = block.stock
        values = { n: getattr(submission, stock.field_name(n))
                   for n in stock.widget_names() }
        return stock.render(ref.field, **values)
    return ''

@register.simple_tag
def dereference_item_field(ref, item):
    return getattr(item, ref.name) or ''

@register.simple_tag
def dereference_item_file(item):
    return item._file, item._filemeta

@register.filter
def items(val): # dict.items in template doesn't work if 'items' is a key
    return val.items()

@register.filter
def join_ids(values, char):
    return char.join([ str(v.pk) for v in values ])

@register.filter
def underscore(obj, name):
    attr = getattr(obj, '_' + name)
    if callable(attr): return attr()
    return attr

@register.filter
def language_label(lang):
    if lang == 'und': return 'unknown language'
    return lang # TODO

@register.filter
def subtitle_url(item, lang):
    return item._artifact_url('subtitles_' + lang)



# Generated by Django 4.0.3 on 2022-04-06 20:24

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('reviewpanel', '0005_presentation_options'),
    ]

    operations = [
        migrations.AlterField(
            model_name='cohortmember',
            name='object_id',
            field=models.UUIDField(db_index=True),
        ),
        migrations.AlterField(
            model_name='score',
            name='object_id',
            field=models.UUIDField(db_index=True),
        ),
        migrations.CreateModel(
            name='Metric',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.SlugField(allow_unicode=True, max_length=20)),
                ('type', models.CharField(choices=[('count', 'count'), ('avg', 'average'), ('stddev', 'standard deviation'), ('max', 'maximum'), ('min', 'minimum')], default='count', max_length=16)),
                ('count_value', models.PositiveIntegerField(blank=True, null=True)),
                ('count_is_divisor', models.BooleanField(default=False)),
                ('boolean_invert', models.BooleanField(default=False)),
                ('position', models.PositiveIntegerField(default=1)),
                ('admin_enabled', models.BooleanField(default=True)),
                ('panelist_enabled', models.BooleanField(default=False)),
                ('cohort', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='reviewpanel.cohort')),
                ('input', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='inputs', related_query_name='input', to='reviewpanel.input')),
            ],
            options={
                'ordering': ['input', 'position', 'pk'],
            },
        ),
        migrations.AddConstraint(
            model_name='metric',
            constraint=models.UniqueConstraint(fields=('input', 'name'), name='unique_metric_name'),
        ),
    ]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['reviewpanel', 'reviewpanel.migrations', 'reviewpanel.templatetags']

package_data = \
{'': ['*'],
 'reviewpanel': ['static/css/*',
                 'static/img/rp_icons/*',
                 'static/js/*',
                 'templates/admin/reviewpanel/*',
                 'templates/admin/reviewpanel/cohort/*',
                 'templates/reviewpanel/*']}

install_requires = \
['formative>=0.9.6']

extras_require = \
{':python_version >= "3.8" and python_version < "3.9"': ['backports.zoneinfo']}

entry_points = \
{'formative.plugin': ['reviewpanel = reviewpanel:FormativePluginMeta']}

setup_kwargs = {
    'name': 'reviewpanel',
    'version': '0.8.1',
    'description': 'Formative plugin for reviewing and scoring applicant submissions',
    'long_description': 'A [Formative](https://github.com/johncronan/formative) plugin for reviewing\nand scoring applicant submissions, **reviewpanel** allows a group of people to\nevaluate the submissions on a Formative form. You can create cohorts and\npresentation formats, set up panelist logins, define metrics, and\ndetermine scores.\n',
    'author': 'John Kyle Cronan',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/johncronan/reviewpanel',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from django.db.models import Subquery
from django.http import HttpResponse
import pyexcel
from reportlab.pdfgen import canvas as pdfgen_canvas
from reportlab.lib import units, styles
from reportlab import platypus
from urllib.parse import quote
from itertools import groupby
import re

from formative.utils import TabularExport
from .models import Score, Input, Metric
from .templatetags.submission import dereference_block


class MetricsTabularExport(TabularExport):
    def __init__(self, program_form, queryset, **kwargs):
        super().__init__(program_form, queryset, **kwargs)
        self.metrics, self.inputs = [], []
        
        for name in self.args:
            if not self.args[name]: continue
            if name.startswith('metric_'): self.metrics.append(name)
            elif name.startswith('input_') and self.args[name][0] != 'no':
                self.inputs.append(name[len('input_'):])
        
        if self.inputs:
            self.text_scores = {}
            qs = Score.objects.filter(object_id__in=queryset.values('pk'))
            scores = qs.filter(input__name__in=self.inputs)
            for score in scores.order_by('created'):
                app = self.text_scores.setdefault(score.object_id, [])
                app.append(score.text)
    
    def header_row(self):
        row = super().header_row()
        
        for name in self.metrics: row.append(name)
        for name in self.inputs: row.append(name)
        return row
    
    def data_row(self, submission, sub_items):
        row = super().data_row(submission, sub_items)
        
        def field_val(item, field):
            val = getattr(item, field)
            if val is None: return ''
            return val
        
        for name in self.metrics: row.append(field_val(submission, name))
        for name in self.inputs:
            vals = []
            if submission.pk in self.text_scores:
                vals = self.text_scores[submission.pk]
            row.append("\n".join(vals))
        
        return row


class CombinedTabularExport:
    def __init__(self, queryset, **kwargs):
        self.args, self.order_by = kwargs, None
    
    def book(self, queryset):
        ret, scores = {}, Score.objects.all()

        for form in queryset:
            qs = form.model.objects.filter(_submitted__isnull=False)
            if self.order_by: qs = qs.order_by(self.order_by)
            
            args = {}
            for block in form.submission_blocks():
                args['block_' + block.name] = True
            
            collections = {}
            for block in form.collections():
                col = collections.setdefault(block.name, [{}, False, False])
                for field in block.collection_fields():
                    col[0][f'cfield_{block.name}.{field}'] = True
                if block.fixed: col[2] = True
                if block.has_file: col[1] = True
            for col_name, (block_args, file, fixed) in collections.items():
                arg = 'collection_' + col_name
                if fixed: args[arg] = self.args['fixed_collections']
                elif file: args[arg] = self.args['file_collections']
                else: args[arg] = self.args['nonfile_collections']
                if args[arg] == 'no': continue
                args.update(block_args)
            
            if self.args['metrics']:
                input_metrics = Metric.objects.filter(admin_enabled=True,
                                                      input__cohort__form=form)
                for metric in input_metrics.distinct():
                    name = f'metric_{metric.input.name}_{metric.name}'
                    annotation = metric.annotation(scores, object_id='pk')
                    qs = qs.annotate(**{name: annotation})
                    args[name] = True
            
            if self.args['text_inputs']:
                inputs = Input.objects.filter(cohort__form=form,
                                              type=Input.InputType.TEXT)
                for input in inputs: args['input_' + input.name] = 'combine'
            
            export = MetricsTabularExport(form, qs, **args)
            ret[form.slug] = export.data(qs)
        return ret
    
    def response_ods(self, filename, queryset):
        book = self.book(queryset)
        
        stream = pyexcel.save_book_as(bookdict=book, dest_file_type='ods')
        mime_type = 'application/vnd.oasis.opendocument.spreadsheet'
        response = HttpResponse(stream, content_type=mime_type)
        
        disp = f"attachment; filename*=UTF-8''" + quote(filename)
        response['Content-Disposition'] = disp
        return response


class PresentationPrintExport:
    def __init__(self, filename, presentation, **kwargs):
        self.filename, self.args = filename, kwargs
        self.presentation = presentation
        self.orientation = kwargs['orientation']
        self.styles = styles.getSampleStyleSheet()
        
        qs = Metric.objects.filter(input__form=presentation.form,
                                   display_values=True).select_related('input')
        vals_metrics = { f'metric_{m.input.name}_{m.name}': m for m in qs }
        
        self.metrics, self.vals_metrics = {}, {}
        for name in self.args:
            if not self.args[name] or not name.startswith('metric_'): continue
            if name in vals_metrics:
                self.vals_metrics[name] = vals_metrics[name]
            else: self.metrics[name] = True
        
        self.sections = {}
        for sec in presentation.template.sections.filter(h__isnull=False):
            self.sections[sec.name] = (float(sec.x)/100, float(sec.y)/100,
                                       float(sec.w)/100, float(sec.h)/100,
                                       sec.font)
        refs = presentation.references
        self.references = refs.select_related('section').order_by('_rank')
        names = Subquery(self.references.filter(collection='').values('name'))
        self.blocks = { b.name: b for b
                        in presentation.form.blocks.filter(name__in=names) }
    
    def apps_per_page(self, canvas):
        # TODO
        return 2
    
    def font_info(self, font_str):
        strs = font_str.split()
        if len(strs) < 2: return 12, 'sans-serif'
        size_str, family = strs[-2:]
        match = re.match(r'^([0-9.-]+)(.*)$', size_str)
        if not match: return 12, family
        size, unit = match.groups()
        units = {'': 1, 'pt': 1, 'px': 0.75, '%': 0.12, 'em': 12}
        try: value = float(size) * units[unit]
        except KeyError: return 12, family
        return value, family # TODO map family to PDF standard font names
    
    def render_app(self, app, x0, y0, dx, dy, c):
        content = {}
        for ref in self.references:
            if ref.collection: continue # TODO
            section = content.setdefault(ref.section.name, [])
            blabel, ilabel = ref.block_label, ref.inline_label
            val = dereference_block({'blocks': self.blocks}, ref, app)
            section.append(((blabel, ilabel), val))
        
        for sec_name, vals in content.items():
            x, y, w, h, font = self.sections[sec_name]
            style = self.styles['BodyText']
            style.fontSize, _ = self.font_info(font)
            left, top, width, height = x0 + x*dx, y0 - y*dx, w*dx, h*dx
            frame = platypus.Frame(left, top - height, width, height)
            pars = []
            for (blabel, ilabel), val in vals:
                if blabel: pars.append(platypus.Paragraph(blabel, style))
                if ilabel: ilabel += ' '
                par = platypus.Paragraph(ilabel + val, style)
                pars.append(platypus.KeepInFrame(0, 0, [par]))
            frame.addFromList(pars, c)
        
        metrics_name = self.presentation.metrics_section()
        if metrics_name in self.sections:
            x, y, w, h, font = self.sections[metrics_name]
            size, _ = self.font_info(font)
            c.setFontSize(size)
            
            left, top, width, height = x0 + x*dx, y0 - y*dx, w*dx, h*dx
            metrics_h, metric_w = 20, width / len(self.metrics)
            for i, name in enumerate(self.metrics):
                l = left + i * metric_w
                label = name[len('metric_'):].replace('_', ' ')
                val = getattr(app, name)
                if val is None: val = ''
                elif type(val) not in (int, bool): val = f'{val:.3f}'
                c.drawString(l, top, f'{label}: {val}')
            
            vals_h = height / len(self.vals_metrics)
            for i, name in enumerate(self.vals_metrics):
                t = top - metrics_h - i * vals_h
                if app.pk not in self.values[name]: continue
                
                for j, score in enumerate(self.values[name][app.pk]):
                    val, input = str(score.value), self.vals_metrics[name].input
                    if input.type == Input.InputType.TEXT: val = score.text
                    c.drawString(left, t - j*1.3*size, val) # TODO: KeepInFrame
    
    def render_pages(self, queryset, canvas):
        num = self.apps_per_page(canvas)
        height = 11 * units.inch
        incr = height / num
        
        page, i = 1, 0
        for app in queryset:
            y = height - (i % num) * incr - 0
            if i and not (i % num):
                canvas.showPage()
                page += 1
            
            self.render_app(app, 50, y, 8.5 * units.inch - 100, incr, canvas)
            
            if not (i % num):
                canvas.setFontSize(8)
                canvas.drawString(4.25*units.inch, 28, str(page))
            i += 1
        canvas.showPage()
    
    def response(self, queryset):
        self.values = {}
        for name, metric in self.vals_metrics.items():
            subq = Subquery(queryset.values('pk'))
            qs = Score.objects.filter(object_id__in=subq, input=metric.input)
            if metric.cohort: qs = qs.filter(cohort=metric.cohort)
            scores = qs.order_by('object_id', 'created')
            objects = groupby(scores, key=lambda s: s.object_id)
            self.values[name] = { k: list(vals) for k, vals in objects }
        
        response = HttpResponse(content_type='application/pdf')
        canvas = pdfgen_canvas.Canvas(response) # TODO landscape orientation
        self.render_pages(queryset, canvas)
        
        canvas.save()
        disp = f"attachment; filename*=UTF-8''" + quote(self.filename)
        response['Content-Disposition'] = disp
        return response



# Generated by Django 4.0.4 on 2022-04-22 17:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('reviewpanel', '0006_alter_cohortmember_object_id_alter_score_object_id_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='metric',
            name='display_values',
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 4.0.3 on 2022-03-26 22:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('reviewpanel', '0003_alter_cohort_inputs_alter_panel_panelists_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='cohort',
            name='size',
            field=models.PositiveIntegerField(default=0, editable=False),
        ),
    ]



from django.urls import path, include

from .views import ProgramView, FormView, FormInfoView, SubmissionView, \
    SubmissionAdminView, PresentationAdminView


review_patterns = [
    path('<slug:program_slug>/<slug:form_slug>/', include([
        path('', FormView.as_view(), name='form'),
        path('complete', FormInfoView.as_view(
            template_name='reviewpanel/form_complete.html'
        ), name='form_complete'),
        path('submissions/', FormInfoView.as_view(
            template_name='reviewpanel/form_index.html'
        ), name='form_index'),
        path('<uuid:pk>/', SubmissionView.as_view(), name='submission'),
        path('<uuid:pk>/skips', SubmissionView.as_view(skips=True),
             name='submission_skips'),
        path('<uuid:pk>/<int:presentation>/', SubmissionAdminView.as_view(),
             name='submission_admin'),
        path('<int:presentation>/', PresentationAdminView.as_view(),
             name='presentation_admin')
    ])),
    path('<slug:slug>/', ProgramView.as_view(), name='program_index'),
    path('', ProgramView.as_view(show_all=True), name='index')
]

urlpatterns = [
    path('review/', include(review_patterns))
]
